@@ -124,6 +124,9 @@ namespace Beagle.Daemon.BlamQueryable {
 			
 			foreach (Channel channel in collection.Channels)	{
 
+				if (channel.Items == null)
+					continue;
+
 				foreach(Item item in channel.Items) {
 					Indexable indexable = new Indexable (
 						new Uri (channel.Url.Replace ("http://", "rss://") + ";item=" + item.Id));
@@ -196,7 +199,7 @@ namespace Beagle.Daemon.BlamQueryable {
 			Stream stream = new FileStream (filename,
 							FileMode.Open,
 							FileAccess.Read,
-							FileShare.Read);
+							FileShare.ReadWrite);
 			XmlTextReader reader = new XmlTextReader (stream);
 
 			collection = (ChannelCollection) serializer.Deserialize (reader);
