@@ -76,6 +76,7 @@ namespace Beagle.Daemon {
 		private int adds_since_last_optimization = 0;
 		private int removals_since_last_optimization = 0;
 		private bool optimizing = false;
+		private int last_item_count = -1;
 
 		public LuceneDriver (string dir)
 		{
@@ -173,12 +174,16 @@ namespace Beagle.Daemon {
 			// this every time at start-up to help avoid running
 			// out of file descriptors if the beagled were to
 			// crash or be shut down while in a non-optimized state.
-
-			// This creates the index if it doesn't exist
-			IndexWriter writer = new IndexWriter (Store, null, !indexExists);
-			if (indexExists)
-				writer.Optimize ();
-			writer.Close ();
+			// (We do the optimization in a thread since it is I/O bound
+			// and slows down start-up.)
+			if (indexExists) {
+				Thread opt_th = new Thread (new ThreadStart (Optimize));
+				opt_th.Start ();
+			} else {
+				// This creates the index if it doesn't exist
+				IndexWriter writer = new IndexWriter (Store, null, true);
+				writer.Close ();
+			}
 		}
 
 		/////////////////////////////////////////////////////
@@ -205,17 +210,42 @@ namespace Beagle.Daemon {
 			get { return fingerprint; }
 		}
 
+		private IFileAttributesStore fa_store = null;
+
+		public IFileAttributesStore FileAttributesStore {
+			get { return fa_store; }
+			set { fa_store = value; }
+		}
+
 		public bool IsUpToDate (string path)
 		{
-			return ExtendedAttribute.Check (path, fingerprint);
+			if (fa_store == null)
+				return false;
+
+			FileAttributes attr = fa_store.Read (path);
+
+			// FIXME: This check is incomplete
+			return attr != null
+				&& attr.Path == path
+				&& FileSystem.GetLastWriteTime (path) <= attr.LastWriteTime;
 		}
 
 		public void AttachTimestamp (string path, DateTime mtime)
 		{
-			try {
-				ExtendedAttribute.Mark (path, fingerprint, mtime);
-			} catch (Exception ex) {
-				log.Warn ("Couldn't set extended attribute on {0}", path);
+			if (fa_store == null)
+				return;
+
+			FileAttributes attr = new FileAttributes ();
+
+			// FIXME?: we just assume that there won't be collisions.
+			attr.UniqueId = Guid.NewGuid ().ToString ();
+
+			attr.Path = path;
+			attr.LastWriteTime = mtime;
+			attr.LastIndexedTime = DateTime.Now;
+
+			if (! fa_store.Write (attr)) {
+				log.Warn ("Couldn't store file attributes for {0}", path);
 			}
 		}
 
@@ -227,7 +257,7 @@ namespace Beagle.Daemon {
 
 		private Lucene.Net.Store.Directory ourStore = null;
 
-		private Lucene.Net.Store.Directory Store {
+		public Lucene.Net.Store.Directory Store {
 			get { return ourStore; }
 
 			set {
@@ -301,8 +331,7 @@ namespace Beagle.Daemon {
 					// non-transient file ContentUris that
 					// appear to be up-to-date.
 					if (indexable != null
-					    && indexable.ContentUri.IsFile
-					    && ! indexable.DeleteContent
+					    && indexable.IsNonTransient
 					    && IsUpToDate (indexable.ContentUri.LocalPath))
 						continue;
 					
@@ -338,6 +367,7 @@ namespace Beagle.Daemon {
 				reader.Delete (term);
 				++removal_count;
 			}
+			last_item_count = reader.NumDocs ();
 			reader.Close ();
 			watch.Stop ();
 			//Log.Debug ("Step #1: {0} {1} {2}", watch, pending_uris.Count,
@@ -347,8 +377,7 @@ namespace Beagle.Daemon {
 			// Step #2: Cache non-transient content mtimes
 			Hashtable mtimes = new Hashtable ();
 			foreach (Indexable indexable in pending_indexables) {
-				if (indexable.ContentUri.IsFile
-				    && ! indexable.DeleteContent) {
+				if (indexable.IsNonTransient) {
 					try {
 						string path = indexable.ContentUri.LocalPath;
 						mtimes [path] = FileSystem.GetLastWriteTime (path);
@@ -377,6 +406,7 @@ namespace Beagle.Daemon {
 					if (writer == null)
 						writer = new IndexWriter (Store, Analyzer, false);
 					writer.AddDocument (doc);
+					++last_item_count;
 					++add_count;
 				}
 			}
@@ -409,8 +439,8 @@ namespace Beagle.Daemon {
 
 		public bool NeedsOptimize {
 			get { 
-				// FIXME: 107 is a totally arbitrary number.
-				return adds_since_last_optimization + removals_since_last_optimization > 107;
+				// FIXME: 259 is a totally arbitrary number.
+				return adds_since_last_optimization + removals_since_last_optimization > 259;
 			}
 		}
 
@@ -427,6 +457,7 @@ namespace Beagle.Daemon {
 				optimizing = true;
 			}
 
+			Log.Debug ("Optimizing...");
 			IndexWriter writer = new IndexWriter (Store, null, false);
 			writer.Optimize ();
 			writer.Close ();
@@ -442,7 +473,11 @@ namespace Beagle.Daemon {
 
 		public void DoQuery (QueryBody body, IQueryResult result, ICollection list_of_uris, UriFilter uri_filter)
 		{
-			LNS.Searcher searcher = new LNS.IndexSearcher (Store);
+			IndexReader reader = IndexReader.Open (Store);
+			if (last_item_count == -1)
+				last_item_count = reader.NumDocs ();
+
+			LNS.Searcher searcher = new LNS.IndexSearcher (reader);
 			LNS.Query query = ToLuceneQuery (body, list_of_uris);
 
 			LNS.Hits hits = searcher.Search (query);
@@ -470,6 +505,7 @@ namespace Beagle.Daemon {
 			}
 			result.Add (filtered_hits);
 
+			// The call to searcher.Close () closes the IndexReader.
 			searcher.Close ();
 		}
 
@@ -530,6 +566,19 @@ namespace Beagle.Daemon {
 			return DoQueryByUri (new Uri[1] { uri });
 		}
 
+		// We cache the number of documents in the index when readers are
+		// available, so calls to GetItemCount will return immediately
+		// if the driver has been flushed or queried.
+		public int GetItemCount ()
+		{
+			if (last_item_count < 0) {
+				IndexReader reader = IndexReader.Open (Store);
+				last_item_count = reader.NumDocs ();
+				reader.Close ();
+			}
+			return last_item_count;
+		}
+
 
 		///////////////////////////////////////////////////////////////////////////////////////
 
@@ -706,6 +755,17 @@ namespace Beagle.Daemon {
 				luceneQuery.Add (mimeTypeQuery, true, false);
 			}
 
+			// If hit types are specified, we must match one of them.
+			if (body.HasHitTypes) {
+				LNS.BooleanQuery hitTypeQuery = new LNS.BooleanQuery ();
+				foreach (string hitType in body.HitTypes) {
+					Term t = new Term ("Type", hitType);
+					LNS.Query q = new LNS.TermQuery (t);
+					hitTypeQuery.Add (q, false, false);
+				}
+				luceneQuery.Add (hitTypeQuery, true, false);
+			}
+
 			// If a list of Uris is specified, we must match one of them.
 			if (listOfUris != null && listOfUris.Count > 0) {
 				LNS.BooleanQuery uriQuery = new LNS.BooleanQuery ();
